/******************************************************************************
    (c) 1998      P.J. Caulfield          patrick@pandh.demon.co.uk
                  E.M. Serrat             emserrat@geocities.com
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 ******************************************************************************
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netdnet/dn.h>
#include <netdnet/dnetdb.h>
#include <unistd.h>


struct	sockaddr_dn		sockaddr;
struct	accessdata_dn		accessdata;
static  struct	nodeent        *binadr;
static	char			dirname[80];
static	unsigned char		buf[32760];
static	int			buflen,bufptr;
static  int			sockfd;
static  char                   *lasterror;
static  int                     show_bytes = 0;

/* DECnet phase IV limits */
#define MAX_NODE      6
#define MAX_USER     12
#define MAX_PASSWORD 12
#define MAX_ACCOUNT  12
#ifndef FALSE
#define TRUE 1
#define FALSE 0
#endif

static int parse(char *);

/*-------------------------------------------------------------------------*/
static void	dndir_usage(FILE *f)
{
    fprintf(f, "\nUSAGE: dndir [OPTIONS] 'node\"user password\"::filespec'\n\n");
    fprintf(f, "NOTE: The VMS filename really should be in single quotes to\n");
    fprintf(f, "      protect it from the shell\n");

    fprintf(f, "\nOptions:\n");
    fprintf(f, "  -s           show file size\n");
    fprintf(f, "  -d           show file creation date\n");
    fprintf(f, "  -p           show file protection\n");
    fprintf(f, "  -o           show file owner\n");
    fprintf(f, "  -b           show file size in bytes rather than blocks\n");
    fprintf(f, "  -l           long format - show all the above\n");
    fprintf(f, "  -n           don't show header. Useful for shell scripts\n");
    fprintf(f, "  -c           Force single-column output on ttys (default for files)\n");
    fprintf(f, "  -w <size>    width of multi-column output\n");
    fprintf(f, "  -f <size>    maximum width of filename field\n");
    fprintf(f, "  -? -h        display this help message\n");
    fprintf(f, "  -V           show version number\n");
			    
    fprintf(f, "\nExamples:\n\n");
    fprintf(f, " dndir -l 'myvax::*.*'\n");
    fprintf(f, " dndir 'cluster\"patrick thecats\"::disk$users:[cats.pics]*.jpg;'\n");
    fprintf(f, "\n");
}

/*-------------------------------------------------------------------------*/
static void	dap_setup_link(void)
{
    if ((sockfd=socket(AF_DECnet,SOCK_SEQPACKET,DNPROTO_NSP)) == -1) 
    {
	perror("socket");
	exit(-1);
    }

    
    if (setsockopt(sockfd,DNPROTO_NSP,SO_CONACCESS,&accessdata,
		   sizeof(accessdata)) < 0) 
    {
	perror("setsockopt");
	exit(-1);
    }
    
    sockaddr.sdn_family = AF_DECnet;
    sockaddr.sdn_flags	= 0x00;
    sockaddr.sdn_objnum	= 0x11;			/* FAL		*/
    sockaddr.sdn_objnamel	= 0x00;
    memcpy(sockaddr.sdn_add.a_addr, binadr->n_addr,6);

    if (connect(sockfd, (struct sockaddr *)&sockaddr, 
		sizeof(sockaddr)) < 0) 
    {
	perror("socket");
	exit(-1);
    }

}
/*-------------------------------------------------------------------------*/
static void	dap_close_link(void)
{
    if (close(sockfd) < 0)
	perror("close");
}
/*-------------------------------------------------------------------------*/
static void	dap_exchg_config(void)
{
    int		er;
    unsigned char	buf[100];
    unsigned char	confmsg[17] = {
	0x01,	/* Message type 1 = Conf */
	0x00,	/* Flag Field		 */
	0x2E,0x04, /* BufSiz 		 */
	0x07,	/* OS type = VMS	 */
	0x03,   /* RMS-32		 */
	7,0,0,0,0, /* Version 'stuff'     */
	0xA2,	/* Sequential Org Support*/
	0x80,
	0xB0,   /* Message Blocking. OK  */
	0xFE,   /* Supp Extended Attrib  */
	0xA8,   /* Supp. Seq Access/Del  */
	0x7C	/* WildCrds/Ren/Seg Msgs */
    };
    
    if ((er=write(sockfd,confmsg,sizeof(confmsg))) < 0)
    {
	perror("Send DAP config msg");
	exit(-1);
    }			      
    if ((er=read(sockfd,buf,sizeof(buf))) < 0)
    {
	perror("Recv DAP config msg");
	exit(-1);
    }			      
    /* Assume that Peer Supports the minimal characteristis we
       setup here
    */
    if (buf[0] == 0x09)
    {
	printf("Local DAP configuration rejected by remote node\n");
	exit(-1);
    }
}
/*-------------------------------------------------------------------------*/
static	void	dap_directory_lookup(void)
{
    unsigned char	part1[4] = {0x03,		/* Access TYPE	   */
				    0x00,		/* FLAGS	   */
				    0x06,		/* Directory Func  */
				    0x01 		/* I/O err is fatal*/
    }; 
    unsigned char	part2[4] = {0x00,		/* Default Acc GET */
				    0x00,		/* Share Ops GET   */
				    0x31,		/* Main/Date/Protec*/
				    0x00 		/* Password Length */
    }; 
    int	i,er;
    
    for (i = 0; i < 4; i++) buf[i] = part1[i];
    buf[4] = strlen(dirname);
    for (i = 0; i < strlen(dirname); i ++) buf[i+5] = dirname[i];
    for (i = 0; i < 4; i++) buf[i+5+strlen(dirname)] = part2[i];
    
    if ((er=write(sockfd,buf,(strlen(dirname)+9))) < 0)
    {
	perror("Send DAP Directory Access");
	exit(-1);
    }
    buflen = bufptr = 0;
}
/*-------------------------------------------------------------------------*/
static void	dap_send_skip(void)
{	
    unsigned char	contmsg[3] = {0x05,0x00,0x02};
    int		er;
    
    if ( (er=write(sockfd,contmsg,3)) < 0)
    {
	perror("Send DAP skip message");
	exit(-1);
    }
    buflen = bufptr = 0;
}
/*-------------------------------------------------------------------------*/
static void	adjust_bufptr(void)
{
    if (buflen <= bufptr)
    {
	buflen=read(sockfd,buf,sizeof(buf));
	if (buflen < 0) 
	{
	    perror("DAP Read Directory Entry");
	    exit(-1);
	}
	bufptr=0;
    }
}
/*-------------------------------------------------------------------------*/
static int	dap_get_dir_entry(char *name,long *size,char *cdt, 
				  char *owner, char *prot)
#define	BC	buf[bufptr]
#define MC	menu[menuptr] 
{
    int	 retval,len,i,auxptr;
    int  menuptr = 0;
    int  ebk, ffb;
    char menu[6];

    memset(menu, 0, sizeof(menu));
    adjust_bufptr();
    while ((BC != 0x06) && (BC != 0x07) )
    {
	adjust_bufptr();
	switch (BC)
	{
	case 0x09 :
	    bufptr++;
	    if (BC & 2) bufptr++;
	    if ((buf[bufptr] == 0x30) && 
		(buf[bufptr+1] == 0x40) )
	{
	    printf("%s %s\n",name,
		   "(Locked by another user)");
	    dap_send_skip();
	    break;
	}
	else return -BC;
	case 0x0F :
	    adjust_bufptr();
	    bufptr++; if (BC & 2) bufptr++;
	    bufptr ++;
	    retval = BC;
	    bufptr += 1;
	    memset(name,0,80);
	    memcpy(name,&buf[bufptr+1],BC);
	    bufptr += (BC + 1);
	    if ((retval == 0x08) || (retval == 0x04))
		return retval;
	    break;
	case 0x02 :
	    adjust_bufptr();
	    bufptr++;
	    if (BC & 2)
	    {
		bufptr++;
		len = BC+1;
	    }
	    else
		len = buflen;
	    auxptr = bufptr;

	    i=0; bufptr++;
	     
	    while (BC & 0x80)
	    {
		menu[i++] = BC;
		bufptr++;
	    }
	    menu[i++] = BC;
	    
	    bufptr += 1;
	    if (MC & 1)	/* DataType */
	    {
		bufptr += 1;
	    }
	    if (MC & 2)	/* ORG		*/
	    {
		bufptr += 1;
	    }
	    if (MC & 4)	/* RFM		*/
	    {
		bufptr += 1;
	    }
	    if (MC & 8)	/* RAT		*/
	    {
		while (BC & 0x80) bufptr++;
		bufptr++;
	    }
	    if (MC & 16)	/* BLS		*/
	    {
		bufptr += 2;
	    }
	    if (MC & 32)	/* MRS		*/
	    {
		bufptr += 2;
	    }
	    *size=0;
	    if (MC & 64)	/* ALQ		*/
	    {
#ifdef __LITTLE_ENDIAN
		memcpy(size,&buf[bufptr+1],BC); 
#else
		swab(size,&buf[bufptr+1],BC); 
#endif
		bufptr += BC+1;
	    }

	    menuptr++;
	    if (MC & 1)	bufptr++;    /* BKS   */
	    if (MC & 2)	bufptr++;    /* FSZ   */
	    if (MC & 4)              /* MRN   */
		bufptr += BC+1;
	    if (MC & 8)              /* RUNSYS */
		bufptr += BC+1;

	    if (MC & 16) bufptr+=2;  /* DEQ   */
	    if (MC & 32)             /* FOP   */
	    {
		while (BC & 0x80) bufptr++;
		bufptr++;
	    }
	    if (MC & 64) bufptr++;   /* BSZ   */
	    
	    menuptr++;
	    if (MC & 1)             /* DEV   */
	    {
		while (BC & 0x80) bufptr++;
		bufptr++;
	    }
	    if (MC & 2)             /* SDC   */
	    {
		while (BC & 0x80) bufptr++;
		bufptr++;
	    }
	    if (MC & 4) bufptr+=2;  /* LRL   */
	    if (MC & 8)             /* HBK   */
	    {
		bufptr += BC+1;
	    }
	    ebk = 0;
	    if (MC & 16)            /* EBK   */
	    {
#ifdef __LITTLE_ENDIAN
		memcpy(&ebk,&buf[bufptr+1],BC); 
#else
		swab(*ebk,&buf[bufptr+1],BC); 
#endif
		bufptr += BC+1;
	    }
	    ffb = 0;
	    if (MC & 32)            /* FFB   */
	    {
		ffb = *(unsigned short *)(buf+bufptr);
		bufptr += 2;
	    }

	    if (show_bytes)
	    {
		if (ebk)
		    *size = ((ebk-1)*512 + ffb);
		else
		    *size = 0;
	    }

	    bufptr = auxptr + len;
	    break;

	case 0x0D :
	    adjust_bufptr();
	    bufptr++; if (BC & 2) bufptr++;
	    len = BC+1;
	    memcpy(cdt,&buf[bufptr+2],18);
	    cdt[18] = '\0';
	    bufptr += len;
	    break;
	case 0x0E :
	    adjust_bufptr();
	    bufptr++; if (BC & 2) bufptr++;
	    bufptr += 2;
	    memset(owner,0,20);
	    memcpy(owner,&buf[bufptr+1],BC);
	    bufptr += (BC + 1);
	    len = 0;
	    len += sprintf(prot+len,"(");
	    for (i=0; i < 4; i++)
	    {
		if ((BC & 0x01)==0) len+=sprintf(prot+len,"R");
		if ((BC & 0x02)==0) len+=sprintf(prot+len,"W");
		if ((BC & 0x04)==0) len+=sprintf(prot+len,"E");
		if ((BC & 0x08)==0) len+=sprintf(prot+len,"D");
		if (i != 3) len+=sprintf(prot+len,",");
		bufptr += 1;
	    }
	    len += sprintf(prot+len,")");
	    return retval;
	default: 
            bufptr += 1;
            adjust_bufptr();
	}
	adjust_bufptr();
    }

    if (BC == 0x06)
    {
	bufptr += 1;
	if (BC == 0x02)
	    bufptr += 2;
	else    bufptr += 1;
	return 0x01;
    }
    if (BC == 0x07)
    {
        return 0x00; //ACCOMP
    }
    return -BC;
}
/*-------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
    int	    opt,retval;
    char    name[80],cdt[25],owner[20],prot[22];
    long    size;
    int     show_date = 0;
    int     show_size = 0;
    int     show_owner = 0;
    int     show_protection = 0;
    int     show_total = 0;
    int     show_header = 1;
    int     filename_width = 22;
    int     single_column = 0;
    int     term_width;
    int     entries_per_line;
    int     printed;
    int     just_shown_header = 0;
    unsigned long total = 0;

    /*	Zero Structure	AccessData			*/
    memset(&accessdata,0,sizeof(accessdata));

    if (argc < 2)
    {
	dndir_usage(stdout);
	exit(0);
    }

    /* Get the terminal width and use it for the default */
    if (isatty(STDOUT_FILENO))
    {
	struct winsize w;
	int st;

	st = ioctl(STDOUT_FILENO, TIOCGWINSZ, &w);
	if (st)
        {
	    perror("ioctl on stdout");
	    term_width = 80;
        }
	else
	{
	    term_width = w.ws_col;
	}
	if (!term_width) term_width = 80;
    }
    else
    {
	single_column = 1; /* Output to a file */
	term_width = 132;
    }

    
/* Get command-line options */
    opterr = 0;
    optind = 0;
    while ((opt=getopt(argc,argv,"?hvVpndlostbw:f:")) != EOF)
    {
	switch(opt)
	{
	case 'h': 
	    dndir_usage(stdout);
	    exit(1);

	case '?':
	    dndir_usage(stderr);
	    exit(1);

	case 'p':
	    show_protection++;
	    single_column++;
	    break;
	    
	case 'o':	   
	    show_owner++;
	    single_column++;
	    break;

	case 'c':	   
	    single_column++;
	    break;
	    
	case 'b':
	    show_bytes++;
	    break;
	    
	case 's':
	    show_size++;
	    single_column++;
	    break;
	    
	case 'd':
	    show_date++;
	    single_column++;
	    break;
	    	    
	case 'l':
	    show_date++;
	    show_size++;
	    show_protection++;
	    show_owner++;
	    single_column++;
	    show_bytes++;
	    break;

	case 'n':
	    show_header=0;
	    break;

	case 't':
	    show_total++;
	    break;

	case 'w':
	    term_width=atoi(optarg);
	    break;

	case 'f':
	    filename_width=atoi(optarg);
	    break;

	case 'V':
	    printf("\ndndir from dntools version %s\n\n", VERSION);
	    exit(1);
	    break;
	}
    }

    /* Parse the filename into its parts */
    if (!parse(argv[optind]))
    {
	fprintf(stderr, "%s\n", lasterror);
	exit(2);
    }

    /* Work out the width of the printout */
    entries_per_line = term_width / (filename_width+2);
    printed = 0;
    
    dap_setup_link();			/* Setup link for DAP	*/

    dap_exchg_config();

    dap_directory_lookup();		/* Look up for directory */
    
    /* Loop through the files we find */
    while ((retval=dap_get_dir_entry(name,&size,cdt,owner,prot)) > 0)
    {
	if (retval == 0x08 && show_header)
        {
	    printf("\nDirectory of %s",name);
            just_shown_header = 1;
        }
	if (retval == 0x04 && show_header)
	{
	    if (!single_column && !just_shown_header)
            {
	        printf("\n\n");  // space out directories.
                just_shown_header = 0;
            }
	    printf("%s\n\n",name);
	}
	if (retval == 0x02)
	{
	    printf("%-*s  ", filename_width, name);
	    if (single_column)
	    {
		if (strlen(name) >= filename_width &&
		    (show_protection || show_date ||
		    show_owner || show_date || show_size))
		{
		    printf("\n%-*s  ",filename_width, "");
		}
		if (show_size)
		    printf("%7ld ", size);
		if (show_date)
		    printf("%s ", cdt);
		if (show_owner)
		    printf("%s ", owner);
		if (show_protection)
		    printf("%s", prot);
		printf("\n");
	    }
	    else
	    {
		if (++printed >= entries_per_line)
		{
		    printf("\n");
		    printed = 0;
		}
	    }
	    if (show_total) total += size;
	}
    }
    if (retval == -9) printf("Error opening %s\n",dirname);
    dap_close_link();

    if (!single_column) printf("\n");

    if (show_total)
    {
	if (show_bytes)
	{
	    printf("\n%-*s  %7ld\n\n", filename_width, "Total bytes", total);
	}
	else
	{
	    printf("\n%-*s  %7ld\n\n", filename_width, "Total blocks", total);
	}
    }
    return 0;
}
/*-------------------------------------------------------------------------*/
/* Parse the filename into its component parts */
static int parse(char *fname)
{
    enum  {NODE, USER, PASSWORD, ACCOUNT, NAME, FINISHED} state;
    int   n0=0;
    int   n1=0;
    char  node[MAX_NODE+1];
    char *nodenum;
    char *local_user;
    char  lastchar;
    
    state = NODE; /* Node is mandatory */

    while (state != FINISHED)
    {
	switch (state)
	{
	case NODE:
	    if (fname[n0] != ':' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_NODE || 
		    fname[n0] == ' ' || fname[n0] == '\n')
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		node[n1++] = fname[n0++];
	    }
	    else
	    {
		node[n1] = '\0';
		n1 = 0;
		if (fname[n0] == '\"')
		{
		    n0++;
		    state = USER;
		}
		else
		{
		    n0 += 2;
		    state = NAME;
		}
	    }
	    break;

	case USER:
	    if (fname[n0] != ' ' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_USER)
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		accessdata.acc_user[n1++] = fname[n0++];
	    }
	    else
	    {
		accessdata.acc_user[n1] = '\0';
		n1 = 0;
		if (fname[n0] == ' ')
		{
		    state = PASSWORD;
		    n0++;
		}
		else /* Must be a quote */
		{
		    state = NAME;
		    /* Check for :: */
		    n0 += 3;
		}
	    }
	    break;

	case PASSWORD:
	    if (fname[n0] != ' ' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_PASSWORD)
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		accessdata.acc_pass[n1++] = fname[n0++];
	    }
	    else
	    {
		accessdata.acc_pass[n1] = '\0';
		n1 = 0;
		if (fname[n0] == ' ')
		{
		    n0++;
		    state = ACCOUNT;
		}
		else /* Must be a quote */
		{
		    state = NAME;
		    /* Check for :: */
		    n0 += 3;
		}
	    }
	    break;

	case ACCOUNT:
	    if (fname[n0] != '\'' && fname[n0] != '\"')
	    {
		if (n1 >= MAX_ACCOUNT)
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		accessdata.acc_acc[n1++] = fname[n0++];
	    }
	    else
	    {
		accessdata.acc_acc[n1] = '\0';;
		state = NAME;
		n1 = 0;
		/* Check for :: */
		n0 += 3;
	    }
	    break;

	case NAME:
	    strcpy(dirname, fname+n0);
	    state = FINISHED;
	    break;

	case FINISHED: // To keep the compiler happy
	    break;
	} /* switch */
    }

/* tail end validation */
    binadr = getnodebyname(node);
    if (!binadr)
    {
	lasterror = "Unknown or invalid node name ";
	return FALSE;
    }

    /* Default to a full wildcard if no filename is given. This will cope with
       the situation where no directory is provided (node::) and where a
       directory (node::[patrick]) is provided. Logical names must end with
       a colon for this to happen or we will not be able to distinguish it
       from a filename */
    lastchar = dirname[strlen(dirname)-1];
    if (lastchar == ']' || lastchar == ':' || dirname[0] == '\0')
	strcat(dirname, "*.*;*");


    // Try very hard to get the local username
    local_user = getlogin();
    if (!local_user || local_user == (char *)0xffffffff)
      local_user = getenv("LOGNAME");
    if (!local_user) local_user = getenv("USER");
    if (local_user)
    {
        int i;

	strcpy((char *)accessdata.acc_acc, local_user);
	accessdata.acc_accl = strlen((char *)accessdata.acc_acc);
        for (i=0; i<accessdata.acc_accl; i++)
            accessdata.acc_acc[i] = toupper(accessdata.acc_acc[i]);
    }
    else
        accessdata.acc_acc[0] = '\0';


    /* Complete the accessdata structure */
    accessdata.acc_userl = strlen(accessdata.acc_user);
    accessdata.acc_passl = strlen(accessdata.acc_pass);
    accessdata.acc_accl  = strlen(accessdata.acc_acc);
    return TRUE;
}

