/******************************************************************************
    (c) 1998 P.J. Caulfield               patrick@pandh.demon.co.uk
                                          pcaulfield@cix.co.uk
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 ******************************************************************************
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/errno.h>
#include <netdnet/dn.h>
#include <netdnet/dnetdb.h>
#include "file.h"
#include "dnetfile.h"

// Assume the file name is a DECnet name if it has two consecutive colons in it
bool dnetfile::isMine(char *name)
{
    if (strstr(name, "::"))
	return TRUE;
    else
	return FALSE;
}

// Constructor. Parse the name
dnetfile::dnetfile(char *name, int verbosity)
{
    isOpen = FALSE;
    verbose = verbosity;
    lasterror = NULL; // parse() will set this if there is a problem.
    strcpy(fname, name);
    parse();
}

dnetfile::~dnetfile()
{
    dap_close_link();
}

int dnetfile::setup_link(unsigned int bufsize, int rfm, int rat, int xfer_mode)
{
    char *local_user;
    
// If there was a parse error in the file name then fail here
    if (lasterror)
	return -1;

// Save these for later
    user_rfm = rfm;
    user_rat = rat;
    transfer_mode = xfer_mode;
    user_bufsize = bufsize;
    
// Fill in the access structure
    strcpy((char *)accessdata.acc_user, user);
    accessdata.acc_userl = strlen(user);

    // Try very hard to get the local username
    local_user = getlogin();
    if (!local_user || local_user == (char *)0xffffffff)
        local_user = getenv("LOGNAME");
    if (!local_user) local_user = getenv("USER");
    if (local_user)
    {
	strcpy((char *)accessdata.acc_acc, local_user);
	accessdata.acc_accl = strlen((char *)accessdata.acc_acc);
        for (int i=0; i<accessdata.acc_accl; i++)
            accessdata.acc_acc[i] = toupper(accessdata.acc_acc[i]);
	
        if (verbose > 1) printf("Account for DECnet proxy is %s (%d)\n",
				accessdata.acc_acc, accessdata.acc_accl);
    }
    else
        accessdata.acc_accl = 0;
    strcpy((char *)accessdata.acc_pass,password);
    accessdata.acc_passl = strlen(password);

    strcpy(filname, name);

// Resolve the node name
    binadr = getnodebyname(node);
    if (binadr == NULL) 
    {
	lasterror = "Unknown host name";
	return -1;
    }
	
    // Set up the link
    if (!dap_setup_link())
	return dap_exchg_config();
    else
	return -1;
}

// Open a file for writing
int dnetfile::open(char *filename,char *mode)
{
    // Add the remote end's directory spec to the filename.
    strcpy(filname, name);
    strcat(filname, filename);
    return open(mode);
}

// Open a file already named 
int dnetfile::open(char *mode)
{
    int real_rfm, real_rat;
    int status;

    switch (mode[0])
    {
    case 'r':
	writing = FALSE;
	break;

    case 'a':   // Not supported yet (if ever!)
	lasterror ="append unsupported";
	return -1;
	break;

    case 'w':
	writing = TRUE;
	break;
    }

    if (!isOpen) // Open connection for the first time
    {
	file_fsz = 0;
	if (writing)
	{
	    status = dap_send_attributes();
	    
	    // Send default file name
	    if (wildcard)
	    {
		status = dap_send_name();
		if (status) return status;
	    }
	    status = dap_send_access();
	    if (status) return status;

	    status = dap_get_file_entry(&real_rfm, &real_rat);
	    if (status) return status;
#ifdef PJC
	    if (!strncmp(filname+strlen(filname)-6, ".DIR;1", 6))
	    {
	        printf("PJC: Skipping directory (in open())\n");
	        dap_send_skip();
		status = dap_get_file_entry(&real_rfm, &real_rat);
		if (status) return status;
	    }
#endif
	}
	else
	{
	    status = dap_send_access();
	    if (status) return status;
	}

	if (!writing)
	{
	    status = dap_get_file_entry(&real_rfm, &real_rat);
	    if (status) return status;
	    if (verbose > 1) printf("File attributes: rfm: %d, rat: %d\n",
				    real_rfm, real_rat);

// Use the file's real attributes if the user just wants defaults.
	    if (user_rfm == RFM_DEFAULT)  file_rfm = real_rfm;
	    if (user_rat == RAT_DEFAULT)  file_rat = real_rat;
	    
// Get the file's real name in case the user specified a wildcard
	    strcpy(name, filname);
	}

	isOpen = TRUE;
	ateof = FALSE;
    }
    else if (writing) // new file to create on already open link
    {
        dap_send_attributes();
        dap_send_access();
	status = dap_get_file_entry(&real_rfm, &real_rat);
    }
    status = dap_send_connect();
    if (status) return status;

    status = dap_send_get_or_put();
    return status;
}

// Close the file but leave the link open in case there are any more to
// read/write
int dnetfile::close()
{
    return dap_send_accomp();
}

// Read a block or a record
int dnetfile::read(char *buf, int len)
{
    int retlen = dap_get_record(buf, len);
    if (retlen < 0) return retlen; // Empty record.

// If the file has implied carriage control then add an LF to the end of the
// line
    if ((file_rfm != RFM_STMLF) && 
	(file_rat == RAT_CR || file_rat == RAT_PRN))
	buf[retlen++] = '\n';

// Print files have a two-byte header indicating the line length.
    if (file_rat == RAT_PRN) 
    {
	memmove(buf, buf+file_fsz, retlen-file_fsz);
	retlen -= file_fsz;
    }

// FORTRAN files have a leading character that indicates carriage control
    if (file_rat == RAT_FTN)
    {
	switch (buf[0]) 
	{
	  
	case '+': // No new line
	  buf[0] = '\r';
	  break;
	  
	case '1': // Form Feed
	  buf[0] = '\f';
	  break;

	case '0': // Two new lines
	  memmove(buf+1, buf, retlen+1);
	  buf[0] = '\n';
	  buf[1] = '\n';
	  retlen++;
	  break;


	case ' ': // new line
	default:  // Default to a new line. This seems to be what VMS does.
	  buf[0] = '\n';
	  break;
	  break;
	}
    }
    return retlen;
}

// Send a block/record
int dnetfile::write(char *buf, int len)
{
    return dap_put_record(buf, len);
}

/* Get the next filename in a wildcard list */
int dnetfile::next()
{
    int status;
    int real_rfm, real_rat;

    if (!wildcard) 
    {
	return FALSE;
    }

    if ( ((status = dap_get_file_entry(&real_rfm, &real_rat))) )
    {
	if (status != -2) perror("last file");
	return FALSE; // No more files.
    }
    else
    {
#ifdef PJC
      if (!strncmp(filname+strlen(filname)-6, ".DIR;1", 6))
      {
	  printf("PJC: Skipping directory(in next())\n");
	  dap_send_skip();
	  status = dap_get_file_entry(&real_rfm, &real_rat);
	  if (status) return status;
      }
#endif

// Use the file's real attributes if the user just wants defaults.
        if (user_rfm == RFM_DEFAULT)  file_rfm = real_rfm;
        if (user_rat == RAT_DEFAULT)  file_rat = real_rat;
        if (verbose > 1) printf("File attributes: rfm: %d, rat: %d\n",
                                 real_rfm, real_rat);

// Get the file's real name in case the user specified a wildcard
        strcpy(name, filname);

        ateof = FALSE;
  	return TRUE;
    }
}

void dnetfile::perror(char *msg)
{
    if (lasterror)
	fprintf(stderr, "%s: %s\n", msg, lasterror);
    else
	fprintf(stderr, "%s: %s\n", msg, strerror(errno));
}

char *dnetfile::get_basename(int keep_version)
{
    make_basename(keep_version);
    return &basename[0];
}

// Returns TRUE if the target filename looks like a directory
bool dnetfile::isdirectory()
{
    // If the last character of the filename is ']' or ':' then we assume
    // the name is a directory. This means that logical names for
    // directories MUST have a colon at the end.
    if (fname[strlen(fname)-1] == ':' ||
	fname[strlen(fname)-1] == ']')
	return TRUE;
    else
	return FALSE;
}

bool dnetfile::eof()
{
    return ateof;
}

bool dnetfile::iswildcard()
{
    return wildcard;
}

void dnetfile::parse()
{
    enum {NODE, USER, PASSWORD, ACCOUNT, NAME, FINISHED} state;
    int  n0=0;
    int  n1=0;
    user[0] = '\0';
    password[0] = '\0';
    account[0] = '\0';
    
    state = NODE; /* Node is mandatory */

    while (state != FINISHED)
    {
	switch (state)
	{
	case NODE:
	    if (fname[n0] != ':' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_NODE || 
		    fname[n0] == ' ' || fname[n0] == '\n')
		{
		    lasterror = "File name parse error";
		    return;
		}
		node[n1++] = fname[n0++];
	    }
	    else
	    {
		node[n1] = '\0';
		n1 = 0;
		if (fname[n0] == '\"')
		{
		    n0++;
		    state = USER;
		}
		else
		{
		    n0 += 2;
		    state = NAME;
		}
	    }
	    break;

	case USER:
	    if (fname[n0] != ' ' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_USER)
		{
		    lasterror = "File name parse error";
		    return;
		}
		user[n1++] = fname[n0++];
	    }
	    else
	    {
		user[n1] = '\0';
		n1 = 0;
		if (fname[n0] == ' ')
		{
		    state = PASSWORD;
		    n0++;
		}
		else /* Must be a quote */
		{
		    state = NAME;
		    /* Check for :: */
		    n0 += 3;
		}
	    }
	    break;

	case PASSWORD:
	    if (fname[n0] != ' ' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_PASSWORD)
		{
		    lasterror = "File name parse error";
		    return;
		}
		password[n1++] = fname[n0++];
	    }
	    else
	    {
		password[n1] = '\0';
		n1 = 0;
		if (fname[n0] == ' ')
		{
		    n0++;
		    state = ACCOUNT;
		}
		else /* Must be a quote */
		{
		    state = NAME;
		    /* Check for :: */
		    n0 += 3;
		}
	    }
	    break;

	case ACCOUNT:
	    if (fname[n0] != '\'' && fname[n0] != '\"')
	    {
		if (n1 >= MAX_ACCOUNT)
		{
		    lasterror = "File name parse error";
		    return;
		}
		account[n1++] = fname[n0++];
	    }
	    else
	    {
		account[n1] = '\0';;
		state = NAME;
		n1 = 0;
		/* Check for :: */
		n0 += 3;
	    }
	    break;

	case NAME:
	    strcpy(name, fname+n0);

	    // Is this a wildcard filename?
	    if (strchr(name, '*') || 
		strchr(name, '%'))
		wildcard = TRUE;
	    else
		wildcard = FALSE;

	    state = FINISHED;
	    break;

	case FINISHED: // To keep the compiler happy
	    break;
	} /* switch */
    }
    return;
}

/* Work out the base name so that when we copy VMS->Unix we don't get the
   device and version numbers in the final filename */
void dnetfile::make_basename(int keep_version)
{
    char        *start;
    char        *end;
    unsigned int i;
		    
    /* Find the start of the name */
    start = rindex(name, ']');
    if (!start) start = rindex(name, ':');
    if (!start) start = name-1;
    strcpy(basename, start+1);

    /* Remove a version number */
    if (!keep_version)
    {
	end = rindex(basename, ';');
	if (end) *end = '\0';
    }

    /* make all lower case */
    for (i=0; i < strlen(basename); i++)
    {
	basename[i] = tolower(basename[i]);
    }
}

// Return a pretty-printed version of the file name for confirmation and
// logging
char *dnetfile::get_printname(char *filename)
{
    static char pname[1024];
    
    strcpy(pname, node);
    if (*user) 
    {
	strcat(pname, "\"");
	strcat(pname, user);
	if (*password) strcat(pname, " password");
	if (*account)
	{
	    strcat(pname, " ");
	    strcat(pname, account);
	}
	strcat(pname, "\"");
    }
    strcat(pname, "::");
    strcat(pname, volname);

    if (filename) // Use passed-in filename if there is one
    {
      //        strcat(pname, name);
        strcat(pname, filename);
    }
    else
    {
        strcat(pname, dirname);
        strcat(pname, filname);
    }
    return pname;
}

char *dnetfile::get_printname()
{
    return get_printname(NULL);
}

// Return a string telling the user whether we transferred blocks or
// records or what.
char *dnetfile::get_format_name()
{
    switch (transfer_mode)
    {
    case MODE_DEFAULT:
	switch (file_rat)
	{
	case RAT_NONE:
	    return "blocks";
	    
	default:
	    return "records";
	}
	break;

    case MODE_RECORD:
	return "records";
    case MODE_BLOCK:
	return "blocks";
    }
    return "things";
}


int dnetfile::get_umask()
{
    if (verbose > 1) printf("protection = %o\n", prot);
    return prot;
}

int dnetfile::set_umask(int mask)
{
// TODO Set the protection - this may have to be done at $CREATE time in
// which case dncopy needs rejigging (clang!)
    return 0;
}


// Return the largest buffer size less than 'biggest'
int dnetfile::max_buffersize(int biggest)
{
    return biggest>remote_bufsize?remote_bufsize:biggest;
}
